# -*- coding: utf-8 -*-
"""
Authors  : Sebastian Pucher and Alexandre Dareau
Created  : 2020-01-26
Modified : 2020-02-27
Language : Python 3.6.5

Comments :
This script demonstrates how we computed histograms for the following
paper:
    "Precise Lifetime Measurement of the Cesium 52D5/2 State"
    arXiv:1912.10089
We published this script and the required data on Zenodo (link).
If one runs this script, it processes the data of the file
"fig3_time_tags/fig3_fluorescence_time_tags.dat" and the corresponding
JSON file "fig3_time_tags/fig3_info.json".
After a histogram was generated with numpy.histogram, we plot it with
matplotlib.pyplot.
It is also possible to import the present functions to other scripts.
"""

__author__ = "Sebastian Pucher and Alexandre Dareau"
__copyright__ = "Copyright 2020"
__credits__ = ["S. Pucher", "P. Schneeweiss", "A. Rauschenbeutel", "A. Dareau"]
__license__ = "CC BY-SA 4.0"
__version__ = "1.0"
__maintainer__ = "Sebastian Pucher"
__email__ = "sebastian.pucher@hu-berlin.de"
__created__ = "2020-01-26"

import numpy as np
import os
import json


def print_report(message, report_progress=True):
    """
    Shorthand function used to print report messages when
    'report_progress' is set to True.

    Parameters
    ----------
    message : str
        The message to print.
    report_progress : bool, optional
        Defines if the `message` should be printed.
    """
    if report_progress:
        print(message)


def fast_count_lines(data_path):
    """
    Estimates the number of lines in a .dat file from its size
    (one line is about 9.578986902939 bytes)

    Parameters
    ----------
    data_path : str, path
        File path to the time-tags .dat file.

    Returns
    -------
    n_lines : int
        Number of lines estimate from the file size.
    """
    bytes_per_line = 9.578986902939
    file_size = os.path.getsize(data_path)
    n_lines = int(file_size / bytes_per_line)
    return n_lines


def load_data(file, rows_to_read=1e6):
    """
    Load data from a file.
    Return the next `rows_to_read` time-tags from a .dat file.

    Parameters
    ----------
    file : file_object
        An opened file object corresponding to a time-tag .dat file.
    rows_to_read : int, optional
        Number of lines (time-tags) to read. The default is 1e6.

    Returns
    -------
    data : numpy_array
        A NumPy array containing the time-tags.

    """
    data = []
    for _ in range(int(rows_to_read)):
        line = file.readline()
        if not line:
            # end of file reached, we break and return
            break
        time_tag = float(line)  # in picoseconds
        data.append(time_tag)
    data = np.array(data)
    return data


def compute_histogram(data_path,
                      json_path,
                      dt=5e-9,
                      batch_size=1e6,
                      max_line_number=None,
                      report_progress=False,
                      progress_step_report=10):
    """
    Computes a histogram of time-tags and returns the starting times of
    the bins and the corresponding counts in a bin.

    Parameters
    ----------
    data_path : str, path
        Location of the data file with the time-tags in ps.
    json_path : str, path
        Location of a JSON file with the information of the cycle
        length in us.
    dt : int, optional
        Width of one bin in s.
    batch_size : int, optional
        Quantity of lines to analyze at once.
    max_line_number : int, optional
        Quantity of lines to process
    report_progress : bool, optional
        If True, the current status of the processing of the
        time-tags will be printed in percent.
    progress_step_report : int, optional
        Every `progress_step_report` percent, the current progress will
        be printed if `report_progress is` True.

    Returns
    -------
    bins_start : numpy_array
        Starting times of the bins of the histogram
    count_histogram : numpy_array
        Time-tag counts in a bin.
    """

    # -- prepare arrays for histogram
    # get cycle duration from JSON file
    with open(json_path) as json_file:
        json_data = json.load(json_file)
    t_max = json_data['cycle duration']['value'] * 1e-6
    t_min = 0.
    # convert times to picoseconds
    t_max_ps = t_max * 1e12
    t_min_ps = t_min * 1e12
    dt_ps = dt * 1e12
    n_bin_ps = int((t_max_ps-t_min_ps) / dt_ps)
    # generate arrays for the accumulated histogram
    bins = np.linspace(t_min_ps, t_max_ps, n_bin_ps)
    bins_start = bins[:-1] * 1e-6  # in microseconds
    count_histogram = np.zeros(n_bin_ps - 1, dtype='int32')

    # -- generate histogram
    # define number of lines to be processed
    if max_line_number is not None:
        n_target = max_line_number
    else:
        n_target = fast_count_lines(data_path)
    # process file
    print_report('Start processing of time-tags', report_progress)
    with open(data_path) as data_file:
        n_lines = 0
        last_report = 0
        while True:
            # load next batch from data file
            data = load_data(data_file, batch_size)
            if data.size == 0:
                break
            # compute histogram of the loaded sub-dataset
            new_histogram, _ = np.histogram(data, bins)
            # add the new histogram to the "count_histogram" array
            count_histogram = np.add(count_histogram, new_histogram)
            # update current line number
            n_lines += len(data)
            # break if max number of lines reached
            if max_line_number is not None and n_lines >= max_line_number:
                break
            if report_progress:
                # print status
                current_progress = 100 * n_lines / n_target
                if current_progress >= last_report + progress_step_report:
                    message = ' + progress %i %%' % current_progress
                    print_report(message, report_progress)
                    last_report = current_progress
    print_report('End of processing - close file', report_progress)
    return bins_start, count_histogram


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # file paths
    DATA_ROOT = os.path.join('..', 'fig3_time_tags')
    f_path = os.path.join(DATA_ROOT, 'fig3_fluorescence_time_tags.dat')
    info_path = os.path.join(DATA_ROOT, 'fig3_info.json')

    # process files
    times, counts = compute_histogram(f_path,
                                      info_path,
                                      dt=5e-9,
                                      batch_size=1e4,
                                      max_line_number=None,
                                      report_progress=True,
                                      progress_step_report=15)

    # plot the histogram
    plt.figure()
    plt.plot(times, counts)
    plt.xlabel('time t (us)')
    plt.ylabel('fluorescence photon counts')
    plt.show()
