# -*- coding: utf-8 -*-
"""
Authors  : Sebastian Pucher and Alexandre Dareau
Created  : 2020-01-26
Modified : 2020-02-27
Language : Python 3.6.5

Comments :
This script demonstrates how we fit our measured data in the following
paper:
    "Precise Lifetime Measurement of the Cesium 52D5/2 State"
    arXiv:1912.10089
We published this script and the required data on Zenodo (link).
If one runs this script, it processes the data of the file
"histograms/fig3/fig3_fluorescence_histo.npz" and the corresponding
JSON file "histograms/fig3/fig3_info.json".
For the fitting, we use the lmfit library
(https://lmfit.github.io/lmfit-py/).
It is also possible to import the present functions to other scripts.
"""

__author__ = "Sebastian Pucher and Alexandre Dareau"
__copyright__ = "Copyright 2020"
__credits__ = ["S. Pucher", "P. Schneeweiss", "A. Rauschenbeutel", "A. Dareau"]
__license__ = "CC BY-SA 4.0"
__version__ = "1.0"
__maintainer__ = "Sebastian Pucher"
__email__ = "sebastian.pucher@hu-berlin.de"
__created__ = "2020-01-26"

import numpy as np
import os
import json
import matplotlib.pyplot as plt
from lmfit.models import ExponentialModel, ConstantModel
from lmfit import Parameters
import lmfit


def load_numpy_arrays(file_path):
    """
    Load data from a .npz file.
    Returns the bins and the counts of the histogram in `file_path`.

    Parameters
    ----------
    file_path : str, path
        Location of the file containing the time and count information
        in NumPy arrays.

    Returns
    -------
    times : numpy_array
        Starting times of the bins of the time-tag counts.
    counts : numpy_array
        Time_tag counts in a bin.
    """
    npz_file = np.load(file_path)
    times = npz_file['times']  # in microseconds
    counts = npz_file['counts']
    return times, counts


def fit_histogram(file_path,
                  json_path,
                  t_fit_start=1.0,
                  t_fit_stop=10.0,
                  print_measurement_info=False,
                  print_fit_report=False,
                  plot_fit=False):
    """
    Fits an exponential decay with an offset to the measurement data.
    We use the lmfit library for the fit and return the exponential
    decay constant (which is the cesium 5D5/2 lifetime) and its error.

    Parameters
    ----------
    file_path : str, path
        Location of the file containing the time and count information
        in NumPy arrays.
    json_path : str, path
        Location of a JSON file with the information of the cycle
        length in microseconds.
    t_fit_start : float, optional
        Start time of the fit.
    t_fit_stop : float, optional
        Stop time of the fit.
    print_measurement_info : bool, optional
        If True, the measurement information from the JSON file is
        printed.
    print_fit_report : bool, optional
        If True, the fit report is printed.
    plot_fit : bool, optional
        If True, the measurement, its fit, and the residuals are
        plotted.
    Returns
    -------
    decay_constant : float
        Exponential decay constant (5D5/2 lifetime) of the fit.
    decay_constant_err : numpy_array
        The standard deviation of the fit variable `decay_constant`.
    """
    # -- get the information of the measurement
    with open(json_path) as json_file:
        # open JSON file
        json_data = json.load(json_file)
    if print_measurement_info:
        # print information about the measurement from the JSON file
        print('= Information about the measurement:')
        print('  + temperature: '
              '%.1f °C' % json_data['mean temperature']['value'])
        print('  + cycle duration: '
              '%i us' % json_data['cycle duration']['value'])
        print('  + tot. duration: '
              '%.1f h' % json_data['total duration']['value'])
        print('  + photons detected: %1.0e' % json_data['photons detected'])
        print('  + photon / cycle: %1.3g' % json_data['photons per cycle'])
        print('  + reference photons detected:'
              ' %1.0e' % json_data['reference photons detected'])
        print('  + reference photons / cycle: '
              '%1.0e' % json_data['reference photons per cycle'])

    # -- prepare arrays for the fit
    # get cycle duration from JSON file
    cycle_duration = json_data['cycle duration']['value']  # in microseconds
    # load data from a .npz file
    times, counts = load_numpy_arrays(file_path)
    # shift the measured data, so that t = 0 is the point where we turn
    # off the laser
    times = times - 0.5*cycle_duration
    # extract from the `times` and `counts` arrays the arrays which we
    # will fit
    i_fit_start = np.searchsorted(times, t_fit_start)
    i_fit_stop = np.searchsorted(times, t_fit_stop)
    times_fit = times[i_fit_start:i_fit_stop]
    counts_fit = counts[i_fit_start:i_fit_stop]
    counts_fit_std = np.sqrt(counts_fit)

    # -- fit the data
    # use a exponential decay with offset as fit model
    fit_model = ExponentialModel() + ConstantModel()
    # Initial guess for the parameters of the fit
    # Free fit parameters: amplitude, offset (c), and decay (lifetime)
    params = Parameters()
    params.add('amplitude',
               value=np.max(counts_fit) - np.min(counts),
               min=0)
    params.add('c', value=np.min(counts))
    params.add('decay', value=1.4, min=1e-10)
    # fit of the histogram
    fit_result = fit_model.fit(counts_fit,
                               params,
                               x=times_fit,
                               weights=np.reciprocal(counts_fit_std))
    # get the lifetime and its error from the fit
    decay_constant = float(fit_result.params['decay'].value)
    decay_constant_err = float(fit_result.params['decay'].stderr)

    # -- print and plot fit results
    if print_fit_report:
        # print fit report with, among others, the information about
        # the reduced chi-squared
        print('= Fit report:')
        print(fit_result.fit_report())
    if plot_fit:
        # plot the measured data, the fit, and the residuals
        # get the y data of the fit
        counts_fitted = fit_result.eval(x=times_fit)
        # get residuals of the fit
        residuals = fit_result.residual
        # prepare subplots
        f, (ax1, ax2) = plt.subplots(2,
                                     gridspec_kw={'height_ratios': [2, 1]},
                                     sharex=True)
        # plot the measured data and the fit
        ax1.plot(times, counts)
        ax1.plot(times_fit, counts_fitted)
        ax1.set_yscale('log')
        ax1.set_ylabel('fluorescence photon counts')
        # plot the residuals
        ax2.scatter(times_fit, residuals, s=2)
        ax2.set_ylim(-4, 4)
        ax2.set_xlim(0, 1.05*t_fit_stop)
        ax2.set_xlabel('time t (us)')
        ax2.set_ylabel('residuals')
        plt.show()
    return decay_constant, decay_constant_err


if __name__ == "__main__":
    # file paths
    DATA_ROOT = os.path.join('..', 'histograms', 'fig3')
    f_path = os.path.join(DATA_ROOT, 'fig3_fluorescence_histo.npz')
    info_path = os.path.join(DATA_ROOT, 'fig3_info.json')
    # process files
    lifetime_5d, lifetime_5d_err = fit_histogram(f_path,
                                                 info_path,
                                                 t_fit_start=0.9,
                                                 t_fit_stop=15.0,
                                                 print_measurement_info=True,
                                                 print_fit_report=True,
                                                 plot_fit=True)
    # print the fit result of the 5D5/2 state lifetime and its 68%
    # confidence interval.
    print('Lifetime of the 5D5/2 state: '
          '{:1.1f}({:1.0f}) ns'.format(1e3*lifetime_5d, 1e4*lifetime_5d_err))
